#!/usr/bin/php52 -n
<?php
/**
 * -*- coding: utf-8 -*-
 *
 * sessionclean.php - a script to cleanup stale PHP sessions
 *
 * Based on sessionclean by Ondřej Surý
 *
 * Copyright 2013-2015 Ondřej Surý <ondrej@sury.org> (original bash script)
 * Copyright 2018-2020 Sergey Dryabzhinsky <sergey.dryabzhinsky@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

// For this php version, and cli name
$PHP="php52";
// Do not stress server, sleep a little, in microsec
$DO_SLEEP=1000;
$PIDOF_BIN="/bin/pidof";
$RENICE_BIN="/usr/bin/renice";
$IONICE_BIN="/usr/bin/ionice";

function is_exe($fpath) {
	return is_file($fpath) && is_executable($fpath);
}

function get_pid($name) {
	global $PIDOF_BIN;
	if (is_exe($PIDOF_BIN)) {
		exec($PIDOF_BIN.' '.$name.' 2>/dev/null', $output, $return_value);
		return $output;
	}
	return array();
}

/**
 * @param array $filenames
 * @param array $ini
 */
function read_configs($filenames, &$ini) {
	$read_ok = array();
	foreach($filenames as $filename) {
		$parsed = parse_ini_file($filename);
		if ($parsed === false) continue;
		$read_ok[] = $filename;
		$ini = array_merge($ini, $parsed);
	}
	return $read_ok;
}

function get_php_session_options($base_php_conf_dir) {

	// Default PHP settings
	$result = array(
		"handler" => "files",
		"path" => "",
		"maxlifetime" => 1440
	);

	$conf_ini = glob($base_php_conf_dir . '/*.ini');
	$conf_ini += glob($base_php_conf_dir . '/conf.d/*.ini');

	if (!$conf_ini) {
		return $result;
	}

	$ini = array();
	$readed = read_configs($conf_ini, $ini);
	if (!$readed) {
		return $result;
	}
	if (isset($ini["session.save_handler"])) {
		$result["handler"] = $ini["session.save_handler"];
	}
	if (isset($ini["session.save_path"])) {
		$result["path"] = $ini["session.save_path"];
	}
	if (isset($ini["session.gc_maxlifetime"])) {
		$result["maxlifetime"] = (int)$ini["session.gc_maxlifetime"];
	}

	return $result;
}

###
# Do not stress server 2
#

if (function_exists("proc_nice")) {
	proc_nice(5);
} else {
	if (is_exe($RENICE_BIN)) {
		system($RENICE_BIN . " -n 5 -p ".getmypid()." 2>/dev/null >/dev/null");
	}
}

if (is_exe($IONICE_BIN)) {
	system($IONICE_BIN . " -c 3 -p ".getmypid()." 2>/dev/null >/dev/null");
}


# Available SAPI config directory name
$SAPIS = array(
	"apache2" => "apache2",
	"cgi" => $PHP."-cgi",
	"fpm" => $PHP."-fpm",
	"cli" => $PHP
);

$sess_sapi = array();
# 1. First - touch all opened files by all SAPIs
# Iterate through all web SAPIs
foreach ($SAPIS as  $sapi_dir => $sapi_proc) {
	if (is_file("/etc/" . $PHP . '/' . $sapi_dir . "/php.ini")) {
		usleep($DO_SLEEP);
		$sess = get_php_session_options("/etc/" . $PHP . '/' . $sapi_dir);
		$sess_sapi[ $sapi_proc ] = $sess;
		if ($sess["handler"] != "files" || !is_dir($sess["path"])) {
			continue;
		}
		$PIDS = get_pid($sapi_proc);
		foreach($PIDS as $pid) {
			$fd_path = "/proc/" . $pid . "/fd";
			if (is_dir($fd_path)) {
				// Find all opened session files in session.save_path dir
				// Maybe too much files to get'em all, read one by one
				if ($handle = opendir($fd_path)) {
					while (false !== ($fn = readdir($handle))) {
						if ($fn == "." || $fn == "..") continue;
						$lnk_path = $fd_path . '/' . $fn;
						if (is_link($lnk_path)) {
							// File may be suddenly removed
							$lnk_point = @readlink($lnk_path);
							// if file deleted it will be prefixed with ' (deleted)' and not match anyway
							if ($lnk_point && fnmatch($sess["path"]."/sess_*", $lnk_point) && strpos($lnk_point, " (deleted)") === false) {
								// Touch file
								@touch($lnk_point);
							}
						}
						usleep($DO_SLEEP);
					}
					closedir($handle);
				}
			}
		}
	}
}

# 2. Second - remove all old session files
# Iterate through all web SAPIs
foreach ($SAPIS as $sapi_dir => $sapi_proc) {
	$sess = @$sess_sapi[ $sapi_proc ];
	if ($sess && is_dir($sess["path"])) {
		usleep($DO_SLEEP);
		// Remove all files older than session.gc_maxlifetime in session.save_path dir
		$old_time = time() - $sess["maxlifetime"];
		// Maybe too much files to get'em all, read one by one
		if ($handle = opendir($sess["path"])) {
			while (false !== ($fn = readdir($handle))) {
				if ($fn == "." || $fn == "..") continue;
				if (fnmatch("sess_*", $fn)) {
					$fn_path = $sess["path"] . '/' . $fn;
					// File may be suddenly removed
					$mtime = @filemtime($fn_path);
					if ($mtime && $mtime < $old_time) {
						@unlink($fn_path);
					}
				}
				usleep($DO_SLEEP);
			}
			closedir($handle);
		}
	}
}
