#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
# sessionclean.py - a script to cleanup stale PHP sessions
#
# Based on sessionclean by Ondřej Surý
#
# Copyright 2013-2015 Ondřej Surý <ondrej@sury.org> (original bash script)
# Copyright 2018 Sergey Dryabzhinsky <sergey.dryabzhinsky@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import subprocess
from time import sleep, time
from fnmatch import fnmatch
import ConfigParser
import StringIO

# For this php version, and cli name
PHP="php52"
# Do not stress server, sleep a little
DO_SLEEP=0.001
PIDOF_BIN="/bin/pidof"
IONICE_BIN="/usr/bin/ionice"

def is_exe(fpath):
	return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

def get_pid(name):
	try:
		if is_exe(PIDOF_BIN):
			return subprocess.check_output([PIDOF_BIN,name]).split()
	except:
		pass
	return []

class PhpIniParser(ConfigParser.ConfigParser):

	def read(self, filenames):
		if isinstance(filenames, ("".__class__, u"".__class__)):
			filenames = [filenames]
		read_ok = []
		for filename in filenames:
			try:
				fp = open(filename)
			except IOError:
				continue
			try:
				self._read(fp, filename)
			except ConfigParser.MissingSectionHeaderError:
				self._MissingSectionHeaderError_read(filename)
			fp.close()
			read_ok.append(filename)
		return read_ok

	def _MissingSectionHeaderError_read(self, filename):
		with open(filename) as cf:
			properties = cf.read()
			configs = "[main]\n%s" % properties
			config_file_stringio = StringIO.StringIO(configs)
			self.readfp(config_file_stringio)


def get_php_session_options(base_php_conf_dir):

	# Default PHP settings
	result = {
		"handler":"files",
		"path":"",
		"maxlifetime":1440
	}

	conf_ini = []
	for root, dirs, files in os.walk(base_php_conf_dir):
		for fn in files:
			if fnmatch(fn, "*.ini"):
				fn_path = os.path.join(root, fn)
				conf_ini.append(fn_path)

	if not len(conf_ini):
		return result

	conf_ini = sorted(conf_ini, key=lambda file: (os.path.dirname(file), os.path.basename(file)))

	config = PhpIniParser(allow_no_value=True)
	readed = config.read(conf_ini)
	if not len(readed):
		return result
	for sect in ("PHP", "Session", "session",):
		if config.has_option(sect, "session.save_handler"):
			result["handler"] = config.get(sect, "session.save_handler")
		if config.has_option(sect, "session.save_path"):
			result["path"] = config.get(sect, "session.save_path")
		if config.has_option(sect, "session.gc_maxlifetime"):
			result["maxlifetime"] = config.getint(sect, "session.gc_maxlifetime")

	return result

###
# Do not stress server 2
#

if hasattr(os, "nice"):
	os.nice(5)

try:
	import psutil
	p = psutil.Process(os.getpid())
	p.ionice(psutil.IOPRIO_CLASS_IDLE)
	IOnice=True
except:
	IOnice=False
	pass
if not IOnice:
	if is_exe(IONICE_BIN):
		subprocess.check_output([IONICE_BIN,"-c","3","-p",str(os.getpid())])


# Available SAPI config directory name
SAPIS={"apache2":"apache2", "cgi":PHP+"-cgi", "fpm":PHP+"-fpm", "cli":PHP}

sess_sapi = {}
# 1. First - touch all opened files by all SAPIs
# Iterate through all web SAPIs
for sapi_dir, sapi_proc in SAPIS.items():
	if os.path.isfile(os.path.join("/etc", PHP, sapi_dir, "php.ini")):
		sleep(DO_SLEEP)
		sess = get_php_session_options(os.path.join("/etc", PHP, sapi_dir))
		sess_sapi[ sapi_proc ] = sess
		if sess["handler"] != "files" or not os.path.isdir(sess["path"]):
			continue
		PIDS = get_pid(sapi_proc)
		for pid in PIDS:
			fd_path = os.path.join("/proc", pid, "fd")
			if os.path.isdir(fd_path):
				# Find all opened session files in session.save_path dir
				for root, dirs, files in os.walk(fd_path):
					for fn in files:
						lnk_path = os.path.join(root, fn)
						if os.path.islink(lnk_path):
							try:
								# File may be suddenly removed
								lnk_point = os.readlink(lnk_path)
								# if file deleted it will be prefixed with ' (deleted)' and not match anyway
								if fnmatch(lnk_point, sess["path"]+"/sess_*") and lnk_point.find(" (deleted)") == -1:
									# Touch file
									os.utime(lnk_point, None)
							except:
								pass
						sleep(DO_SLEEP)

# 2. Second - remove all old session files
# Iterate through all web SAPIs
for sapi_dir, sapi_proc in SAPIS.items():
	sess = sess_sapi.get( sapi_proc )
	if sess and os.path.isdir(sess["path"]):
		sleep(DO_SLEEP)
		# Remove all files older than session.gc_maxlifetime in session.save_path dir
		old_time = time() - sess["maxlifetime"]
		for root, dirs, files in os.walk(sess["path"]):
			for fn in files:
				if fnmatch(fn, "sess_*"):
					fn_path = os.path.join(root, fn)
					try:
						# File may be suddenly removed
						mtime = os.path.getmtime(fn_path)
						if mtime < old_time:
							os.unlink(fn_path)
					except:
						pass
				sleep(DO_SLEEP)
